#include <libce.h>
// per usare le funzioni di libce dedicate alla memoria virtuale
// si deve includere il file vm.h dopo aver incluso libce.h
// I tipi e le funzioni usate sono descritte in
//
//   https://calcolatori.iet.unipi.it/resources/paginazione-libreria.pdf
//
// e nella documentazione della libce.
#include <vm.h>

// scriviamo un programma che permetta di accedere alla variabile x, oltre che
// dall'indirizzo virtuale &x, anche a un altro indirizzo virtuale scelto da
// noi.
int x = 0;

void main()
{
	// Scegliamo un indirizzo virtuale.  Visto che la MMU scompone gli
	// indirizzi virtuali in campi da 9 bit (indici nelle tabelle) e 12 bit
	// (offset), conviene scrivere l'indirizzo in base 8. In questo modo I
	// gli indici si scrivono esattamente su 3 cifre e l'offset su 4.
	//
	//      livello 4  3  2  1  offset
	vaddr v = norm(04443332221110000);
	// L'indirizzo 0444333222110000 non è normalizzato, perché il bit 47
	// vale 1 e dunque i 16 bit più significativi devono essere anch'essi
	// 1. Usiamo la funzione norm() per normalizzarlo.

	// Per poter accedere a x dall'indirizzo virtuale scelto da noi,
	// dobbiamo creare una nuova traduzione dall'indirizzo virtuale
	// all'indirizzo fisico di x. Troviamo intanto l'indirizzo fisico di x:
	paddr px = trasforma(readCR3(), int_cast<vaddr>(&x));
	// Usiamo readCR3() per ottenere la tabella radice del TRIE corrente
	// (creato dal boot loader). L'int_cast<> è necessario perché per il
	// compilatore C++ &x è un puntatore, mentre a noi serve un vaddr, che
	// è un tipo intero.
	// Nota: in questo caso l'indirizzo fisico sarà numericamente uguale
	// all'indirizzo virtuale (&x), in quanto il programma gira all'interno
	// della finestra di memoria creata dal boot loader.

	// Non possiamo, in generale, far corrispondere v a px, perché la MMU
	// ci permette ti trasformare solo il numero di pagina di un indirizzo
	// virtuale, mentre l'offset rimane lo stesso. Quello che possiamo fare
	// è trasformare v nell'indirizzo base del frame in cui px cade, sia
	// base_px. Questo, automaticamente, trasforma (v+o) in (base_px+o) per
	// tutti gli 0 <= o < 4096.  Calcoliamo dunque base_px: è sufficiente
	// azzerare i 12 bit meno significativi di px, cosa che possiamo fare
	// con la maschera 0xFFFFFFFFFFFFF000 = ~0xFFF.
	paddr base_px = px & ~0xFFF;
	// (alternativamente si può usare la funzione base(px, 0))

	// Ora creiamo la traduzione v -> base_px. Il percorso di traduzione di
	// v parte dalla tabella radice (attualmente puntata da cr3) e prosegue
	// per tutti e 4 i livelli, usando gli indici 0444, 0333, 0222 e 0111
	// estratti da v. La parte di TRIE puntata dall'entrata 0444 della
	// tabella di livello 4 è sicuramente assente nel TRIE creato dal boot
	// loader, perché il boot loader ha creato soltanto una finestra di 4
	// GiB, che è interamente contenuta nel sotto-albero puntato
	// dall'entrata 0. Dobbiamo quindi allocare le tabelle di livello 3, 2
	// e 1 per completare il percorso di traduzione di v.  Si ricordi che
	// queste tabelle devono essere allineate naturalmente, cioè il loro
	// indirizzo deve essere multiplo della loro dimensione (4096 byte).
	// Usiamo la funzione alloca_tab() della libreria che le alloca nello
	// heap (creato dal boot loader) avendo cura di rispettare
	// l'allineamento:
	paddr tab4 = readCR3();
	paddr tab3 = alloca_tab();
	paddr tab2 = alloca_tab();
	paddr tab1 = alloca_tab();
	// Nota: stiamo assumento che alloca_tab() abbia sempre successo, cosa
	// sicuramente vera per questo semplice programma. In generale,
	// potrebbe restituire 0 se l'allocazione fallisce.

	// Ora prendiamo dei riferimenti a tutte le entrate che dovremo
	// riempire per creare la nuova traduzione:
	tab_entry& e4 = get_entry(tab4, 0444);
	tab_entry& e3 = get_entry(tab3, 0333);
	tab_entry& e2 = get_entry(tab2, 0222);
	tab_entry& e1 = get_entry(tab1, 0111);
	// Nota: invece di scrivere esplicitamente 0444, 0333 etc, avremmo potuto usare i_tab(v, 4),
	// i_tab(v, 3), etc.

	// Infine riempiamo le entrate. Le tabelle di livello 4, 3 e 2 devono
	// contenere il numero di frame della tabella inferiore, mentre la
	// tabella di livello 1 deve contenere il numero di frame in cui la MMU
	// dovrà tradurre v (base_px nel nostro caso). Sfruttiamo il fatto che,
	// all'interno di ogni entrata, il campo F (numero di frame) si trova
	// nella stessa posizione in cui si trova in un indirizzo fisico,
	// quindi possiamo, per esempio, assegnare direttamente e4=tab3 per
	// scrivere correttamente il numero di frame di tab3 nel campo F di e4,
	// e lo stesso per la traduzione finale: e1=base_px. Inoltre, sia gli
	// indirizzi delle tabelle, sia la traduzione finale hanno i 12 bit
	// meno significativi a 0. Quindi possiamo settare i bit accessori che
	// ci servono con degli OR bit a bit.  Almeno un bit accessorio va
	// settato sicuramente: il bit P, senza il quale la MMU non considera
	// la traduzione valida. Per quanto riguarda gli altri bit:
	//
	// - A e D saranno settati dalla MMU durante gli accessi e normalmente
	//   vanno messi a zero quando si crea una nuova traduzione;
	// - PWT e PCD possiamo lasciarli a zero, in quanto stiamo creando una
	//   traduzione per della normale RAM, per la quale ha senso usare la
	//   cache;
	// - U/S possiamo lasciarlo a 0 (accesso negato da livello utente), in
	//   quanto il programma che stiamo scrivendo gira a livello sistema;
	// - R/W va settato a 1 se vogliamo permettere la scrittura. Per il
	//   momento lo lasciamo a zero.
	//
	e4 = tab3    | BIT_P;
	e3 = tab2    | BIT_P;
	e2 = tab1    | BIT_P;
	e1 = base_px | BIT_P;

	// A questo punto la traduzione è attiva. La variabile x sarà
	// accessibile anche da v + l'offset di x all'interno del frame con
	// base base_px. Calcoliamo questo offset (12 bit meno sinificativi di
	// px):
	natq offset_px = px & 0xFFF;
	// L'indirizzo virtuale alternativo per &x avrà il numero di pagina di
	// v e l'offset offset_px:
	vaddr vx = v | offset_px;
	// Nota: possiamo usare un semplice or perché v ha i 12 bit meno
	// significativi a zero

	// A questo punto, per usare il nostro nuovo indirizzo virtuale dal
	// C++, dobbiamo trasformarlo in un puntatore:
	int *p = ptr_cast<int>(vx);

	// Ora mostriamo che *p punta effettivamente a x, scrivendo in x e poi
	// leggendo *p:
	printf("*p prima dell'assegnamento: %d\n", *p);
	x = 1;
	printf("*p dopo l'assegnamento: %d\n", *p);
	pause();

}
