#include <libce.h>
#include <vm.h>

// Il tab_iter può essere usato per visitare un intero sottoalbero,
// specificando un intervallo di indirizzi. La visita è in ordine depth-first,
// rispettando l'ordine degli indirizzi. Possiamo osservare cosa accade
// visitando parte del TRIE creato dal boot loader.
void main()
{
	// Visitiamo tutto il sottoalbero che si occupa della traduzione degli
	// indirizzi virtuali in [0, 4*MiB). Dovremmo vedere l'iteratore che
	// scende subito dal livello 4 al livello 1 (depth-first), visitando
	// in ordine tutte le 512 entrate della tabella di livello 1, che
	// coprono 2*MiB di indirizzi virtuali.  Terminata la tabella di
	// livello 1, risale alla seconda entrata della tabella di livello 2.
	// In questo caso la visita termina, perché il boot loader usa pagine
	// di livello 2 dalla seconda entrata in poi, e dunque questa entrata
	// copre i rimanenti 2*MiB.
	for (tab_iter it(readCR3(), 0, 4*MiB); it; it.next()) {
		int l = it.get_l();
		tab_entry e = it.get_e();
		paddr tab = it.get_tab();
		// ci facciamo anche dire l'indirizzo più piccolo la cui
		// traduzione passa da questa entrata:
		vaddr v = it.get_v();
		// Questo ciclo stamperà molti messaggi.  Invece di stamparli
		// sul monitor della macchina virtuale, che ha dimensione
		// limitata, li inviamo sul log di sistema, che viene invece
		// mostrato nel terminale da cui abbiamo lanciato QEMU.
		// Notare che flog() aggiunge automaticamente un a-capo alla fine.
		flog(LOG_DEBUG, "indirizzo %6lx: livello %d, tabella %6lx, entrata %6lx",
				v, l, tab, e);
		// Nota: se l'output sembra incompleto, premere ESC per fare terminare
		// QEMU. A questo punto dovrebbe apparire l'output completo. Questo
		// comportamento è dovuto alla "bufferizzazione" dello stdout di QEMU,
		// che lo script `boot` invia, per default, a due altri script che
		// si occupano di colorare i messaggi del log e risolvere eventuali
		// indirizzi in `file:numero-di-riga`. Entrambe le funzioni possono
		// essere disabilitate lanciando
		//
		//   CERAW=1 boot -C
		//
		// In particolare, `CERAW=1` disabilita la risoluzione degli indirizzi
		// e `-C` disabilta la colorazione. In questo modo l'output di QEMU non
		// sarà bufferizzato e non sarà necessario far terminare QEMU per vederlo
		// tutto.
	}
	// Cose da notare nell'output:
	//
	// - La traduzione creata dal boot loader è la traduzione identità: gli
	//   indirizzi virtuali sono uguali agli indirizzi fisici contenuti nelle
	//   tabelle foglia (livello 1 o livello 2 a seconda del bit PS);
	//
	// - la prima entrata della tabella di livello 1 contiene 0 (la pagina
	//   0 non è mappata, per intercettare *nullptr);
	//
	// - L'entrata di livello 2 relativa all'indirizzo 0x200000 (2*MiB) ha
	//   il bit PS settato (bit numero 7).
	//
	// - Le entrate di livello 1 relative agli indirizzi da 0xa0000 e
	//   0xc0000 (escluso) hanno il bit PWT settato (bit numero 4); si tratta
	//   degli indirizzi dedicati alla memoria video;
	//
	// - alcune entrate hanno già i bit A o D settati. Si tratta degli
	//   indirizzi a cui il software ha già acceduto da quando il boot loader
	//   ha creato il TRIE e ha attivato la paginazione; tutto il boot loader
	//   (codice, dati e pila) si trova nelle prime 11 pagine del secondo MiB
	//   (quindi in [0x100000, 0x10b000)), mentre tutto il programma che stiamo
	//   leggendo si trova nel terzo MiB (quindi a partire da 0x200000).
	//
	// - Le entrate relative agli indirizzi 0x2000, 0x4000, 0x6000 e 0x8000
	//   hanno il bit A settato. Questi sono gli indirizzi (fisici) che
	//   corrispondono alle tabelle del TRIE, e i bit A sono stati settati
	//   proprio a causa della visita fatta dal tab_iter (tramite la finestra
	//   stessa, e dunque ad indirizzi virtuali uguali numericamente agli
	//   indirizzi fisici).
	pause();
}
